###
### blunder.tcl: part of Scid.
### Copyright (C) 2006  Pascal Georges
###

######################################################################
### Blunder window: uses a chess engine (Phalanx) in easy mode and
### another engine (for example Crafty) to track blunders

namespace eval ::blunder {}

#set ::engines(list) {}
# analysisCoach(logMax):
#   The maximum number of log message lines to be saved in a log file.
set analysisCoach(logMax) 500

# analysisCoach(log_stdout):
#   Set this to 1 if you want Scid-Engine communication log messages
#   to be echoed to stdout.
set analysisCoach(log_stdout) 0

set level 1500
# default value for considering the engine (Phalanx) blundering
set threshold 0.9
# minimal value for ?!
variable ::blunder::threshold1 0.5
# minimal value for ?
variable ::blunder::threshold2 1.5
# minimal value for ??
variable ::blunder::threshold3 3.0
variable ::blunder::thresholdGameWasDeadAnyway 3.5
set resignValue 4.5
set resignCount 0

set index1 0
set index2 0

set lastblundervalue 0.0
set prev_lastblundervalue 0.0
set blundermissed false
set blunderwarning false
set blunderwarningvalue 0.0
set blundermissedvalue 0.0

set blunderWarningLabel $::tr(Noblunder)
set scoreLabel 0.0

set blunderpending 0
set prev_blunderpending 0
set showblunder 1
set showblundervalue 1
set showblunderfound 1
set showmovevalue 1
set showevaluation 1
set isLimitedAnalysisTime 1
set analysisTime 10
set currentPosHash 0
set lscore {}

set analysisCoach(automove1) 0

# ======================================================================
# resetValues
#   Resets all blunders data.
# ======================================================================
proc ::blunder::resetValues {} {
  
  set ::level 100
  set ::threshold 0.9
  set ::blundermissed false
  set ::blunderwarning false
  set ::blunderwarningvalue 0.0
  set ::lastblundervalue 0.0
  set ::prev_lastblundervalue 0.0
  set ::prev_blunderpending 0
  set ::currentPosHash [sc_pos hash]
  set ::lscore {}
  set resignCount 0
}

# ======================================================================
# resetEngine:
#   Resets all engine-specific data.
# ======================================================================
proc ::blunder::resetEngine {n} {
  
  global analysisCoach
  set analysisCoach(pipe$n) ""             ;# Communication pipe file channel
  set analysisCoach(seen$n) 0              ;# Seen any output from engine yet?
  set analysisCoach(seenEval$n) 0          ;# Seen evaluation line yet?
  set analysisCoach(score$n) 0             ;# Current score in centipawns
  set analysisCoach(moves$n) ""            ;# PV (best line) output from engine
  set analysisCoach(movelist$n) {}         ;# Moves to reach current position
  set analysisCoach(nonStdStart$n) 0       ;# Game has non-standard start
  set analysisCoach(has_analyze$n) 0       ;# Engine has analyze command
  set analysisCoach(has_setboard$n) 0      ;# Engine has setboard command
  set analysisCoach(send_sigint$n) 0       ;# Engine wants INT signal
  set analysisCoach(wants_usermove$n) 0    ;# Engine wants "usermove" before moves
  set analysisCoach(isCrafty$n) 0          ;# Engine appears to be Crafty
  set analysisCoach(wholeSeconds$n) 0      ;# Engine times in seconds not centisec
  set analysisCoach(analyzeMode$n) 0       ;# Scid has started analyze mode
  set analysisCoach(invertScore$n) 1       ;# Score is for side to move, not white
  set analysisCoach(automove$n) 0
  set analysisCoach(automoveThinking$n) 0
  set analysisCoach(automoveTime$n) 2000
  set analysisCoach(lastClicks$n) 0
  set analysisCoach(after$n) ""
  set analysisCoach(log$n) ""              ;# Log file channel
  set analysisCoach(logCount$n) 0          ;# Number of lines sent to log file
  set analysisCoach(wbEngineDetected$n) 0  ;# Is this a special Winboard engine?
}

# ======================================================================
#		::blunder::readEnginesList
#   read the list of already configured engines
# ======================================================================
proc ::blunder::readEnginesList {} {
  catch { source $::engines(file) }
}

# ======================================================================
#		::blunder::config
#   Configure coach games :
#			- Phalanx engine (because it has an 'easy' option)
#			- Coach engine (Crafty is the best)
#			- level of difficulty
# ======================================================================
proc ::blunder::config {} {
  
  global configWin font_Analysis analysisCoachCommand analysisCoach \
      engineCoach1 engineCoach2 level isLimitedAnalysisTime analysisTime index1 index2
  set w ".configWin"
  if {[winfo exists $w]} {
    focus .
    destroy $w
    set configWin 0
    return
  }
  
  # no need to reread the engines list as it is done somewhere else
  #  ::blunder::readEnginesList
  
  toplevel $w
  wm title $w "$::tr(configurecoachgame)"
  
  bind $w <F1> { helpWindow ConfigureCoach }
  setWinLocation $w
  
  frame $w.fexplanation
  frame $w.fengines -relief groove -borderwidth 1
  frame $w.flevel -relief groove -borderwidth 1
  frame $w.flimit -relief groove -borderwidth 1
  frame $w.fbuttons
  
  pack $w.fexplanation $w.fengines $w.flevel $w.flimit $w.fbuttons -side top -fill x
  
  label $w.fexplanation.text -width 50 -height 5 -fg black -bg #b0d1c4f1def9 -justify left -wraplength 300 -font font_Fixed \
      -text $::tr(selectCoachEngines)
  pack $w.fexplanation.text -side top -fill both
  
  frame $w.fengines.left -relief raised -borderwidth 1
  frame $w.fengines.right -relief raised -borderwidth 1
  label $w.fengines.left.l -text $::tr(Phalanxengine)
  label $w.fengines.right.l -text $::tr(Coachengine)
  listbox $w.fengines.left.list -selectmode single -exportselection 0 -yscrollcommand "$w.fengines.left.ybar set"
  listbox $w.fengines.right.list -selectmode single -exportselection 0 -yscrollcommand "$w.fengines.right.ybar set"
  scrollbar $w.fengines.left.ybar -command "$w.fengines.left.list yview"
  pack $w.fengines.left.ybar -side right -fill y
  scrollbar $w.fengines.right.ybar -command "$w.fengines.right.list yview"
  pack $w.fengines.right.ybar -side right -fill y
  
  set i 0
  set index1 0
  set index2 0
  foreach e $::engines(list) {
    set name [lindex $e 0]
    if [ string equal $name $engineCoach1 ] {
      set index1 $i
    }
    if [ string equal $name $engineCoach2 ] {
      set index2 $i
    }
    $w.fengines.left.list insert end $name
    $w.fengines.right.list insert end $name
    incr i
  }
  
  $w.fengines.left.list selection set $index1
  $w.fengines.right.list selection set $index2
  
  pack $w.fengines.left.l $w.fengines.left.list -side top
  pack $w.fengines.right.l $w.fengines.right.list -side top
  pack $w.fengines.left $w.fengines.right -side left -expand yes -fill both
  
  bind $w.fengines.left.list <ButtonRelease-1> {
    set engineCoach1 [ .configWin.fengines.left.list get [ .configWin.fengines.left.list curselection ]]
    set index1 [ .configWin.fengines.left.list curselection ]
  }
  bind $w.fengines.right.list <ButtonRelease-1> {
    set engineCoach2 [ .configWin.fengines.right.list get [ .configWin.fengines.right.list curselection ]]
    set index2 [ .configWin.fengines.right.list curselection ]
  }
  
  scale $w.flevel.difficulty -orient horizontal -from 1200 -to 2200 -length 200 -label $::tr(difficulty)	\
      -variable level -tickinterval 0 -resolution 50
  label $w.flevel.easy -font font_Fixed -text $::tr(easy)
  label $w.flevel.hard -font font_Fixed -text $::tr(hard)
  pack $w.flevel.easy $w.flevel.difficulty $w.flevel.hard -side left -expand yes
  
  # in order to limit CPU usage, limit the time for analysis (this prevents noise on some laptops)
  checkbutton $w.flimit.blimit -text $::tr(limitanalysis) -variable isLimitedAnalysisTime -relief flat
  scale $w.flimit.analysisTime -orient horizontal -from 5 -to 60 -length 200 -label $::tr(seconds) -variable analysisTime -resolution 5
  pack $w.flimit.blimit $w.flimit.analysisTime -side left -expand yes
  
  button $w.fbuttons.close -text $::tr(Play) -command " focus .; destroy $w; ::blunder::play "
  button $w.fbuttons.cancel -textvar ::tr(Cancel) -command "focus .; destroy $w"
  
  pack $w.fbuttons.close $w.fbuttons.cancel -expand yes -side left -padx 20 -pady 2
  
  bind $w <F1> { helpWindow Coach }
  bind $w <Escape> "destroy $w"
  bind $w <Destroy> ""
  bind $w <Configure> "recordWinSize $w"
  wm minsize $w 45 0
}

# ======================================================================
#
#	::blunder::play
#
# ======================================================================
proc ::blunder::play { } {
  global analysisCoach threshold showblunder showblundervalue blunderfound showmovevalue level \
      engineCoach1 engineCoach2 index1 index2
  
  set w ".coachWin"
  if {[winfo exists $w]} {
    focus .
    destroy $w
    set coachWin 0
    return
  }
  
  toplevel $w
  wm title $w "$::tr(coachgame) (Elo $level)"
  
  setWinLocation $w
  
  frame $w.fdisplay -relief groove -borderwidth 1
  frame $w.fthreshold -relief groove -borderwidth 1
  frame $w.finformations -relief groove -borderwidth 1
  frame $w.fbuttons
  pack $w.fdisplay $w.fthreshold $w.finformations $w.fbuttons -side top -expand yes -fill both
  
  checkbutton $w.fdisplay.b1 -text $::tr(showblunderexists) -variable showblunder -relief flat
  checkbutton $w.fdisplay.b2 -text $::tr(showblundervalue) -variable showblundervalue -relief flat
  checkbutton $w.fdisplay.b5 -text $::tr(showscore) -variable showevaluation -relief flat
  pack $w.fdisplay.b1 $w.fdisplay.b2 $w.fdisplay.b5 -expand yes -anchor w
  
  label $w.fthreshold.l -text $::tr(moveblunderthreshold) -wraplength 300 -font font_Fixed
  scale $w.fthreshold.t -orient horizontal -from 0.0 -to 10.0 -length 200 \
      -variable threshold -resolution 0.1 -tickinterval 2.0
  pack $w.fthreshold.l $w.fthreshold.t -side top
  
  label $w.finformations.l1 -textvariable blunderWarningLabel -bg linen
  label $w.finformations.l3 -textvariable scoreLabel -fg WhiteSmoke -bg SlateGray
  pack $w.finformations.l1 $w.finformations.l3 -padx 5 -pady 5 -side top -fill x
  
  button $w.fbuttons.close -textvar ::tr(Abort) \
      -command "focus .; destroy $w ; ::blunder::closeEngine 1; ::blunder::closeEngine 2"
  pack $w.fbuttons.close -expand yes -fill both -padx 20 -pady 2
  
  ::blunder::launchengine $index1 1
  ::blunder::launchengine $index2 2
  ::blunder::resetValues
  updateAnalysisText
  
  bind $w <F1> { helpWindow Coach }
  bind $w <Destroy> "::blunder::closeEngine 1; ::blunder::closeEngine 2"
  bind $w <Escape> "destroy $w"
  wm minsize $w 45 0
  
  # Phalanx engine plays by itself but make sure first that analysis engine is started
  while {!$analysisCoach(has_analyze2) } {
    update
    after 500
  }
  ::blunder::phalanxGo
}

# ======================================================================
#   ::blunder::launchengine
#  - launches both engines
#  - updates values for :
#						blundermissed (boolean), blunderwarning (boolean)
#						blunderwarningvalue (real), blundermissedvalue (real)
#						totalblundersmissed (real), totalblunders (real)
# ======================================================================
proc ::blunder::launchengine {index n} {
  global analysisCoach level
  
  ::blunder::resetEngine $n
  
  set engineData [lindex $::engines(list) $index]
  set analysisName [lindex $engineData 0]
  set analysisCommand [lindex $engineData 1]
  set analysisArgs [lindex $engineData 2]
  set analysisDir [lindex $engineData 3]
  if {$analysisArgs != ""} {
    append analysisCommand " $analysisArgs"
  }
  # turn phalanx book, ponder and learning off, easy on
  if {$n == 1} {
    # convert Elo = 1200 to level 100 up to Elo=2200 to level 0
    set easylevel [expr 100-(100*($level-1200)/(2200-1200))]
    append analysisCommand " -b+ -p- -l- -e $easylevel "
  }
  
  # If the analysis directory is not current dir, cd to it:
  set oldpwd ""
  if {$analysisDir != "."} {
    set oldpwd [pwd]
    catch {cd $analysisDir}
  }
  
  # Try to execute the analysis program:
  if {[catch {set analysisCoach(pipe$n) [open "| $analysisCommand" "r+"]} result]} {
    if {$oldpwd != ""} { catch {cd $oldpwd} }
    tk_messageBox -title "Scid: error starting analysis" \
        -icon warning -type ok \
        -message "Unable to start the program:\n$analysisCommand"
    ::blunder::resetEngine $n
    return
  }
  
  # Return to original dir if necessary:
  if {$oldpwd != ""} { catch {cd $oldpwd} }
  
  # Open log file if applicable:
  set analysisCoach(log$n) ""
  if {$analysisCoach(logMax) > 0} {
    if {! [catch {open [file join $::scidUserDir "coachengine$n.log"] w} log]} {
      set analysisCoach(log$n) $log
      logEngine $n "Scid-Engine communication log file"
      logEngine $n "Engine: $analysisName"
      logEngine $n "Command: $analysisCommand"
      logEngine $n "Date: [clock format [clock seconds]]"
      logEngine $n ""
      logEngine $n "This file was automatically generated by Scid."
      logEngine $n "It is rewritten every time an engine is started in Scid."
      logEngine $n ""
    }
  }
  
  # Configure pipe for line buffering and non-blocking mode:
  fconfigure $analysisCoach(pipe$n) -buffering line -blocking 0
  
  if {$n == 1} {
    fileevent $analysisCoach(pipe$n) readable "::blunder::processInput"
    after 1000 "::blunder::checkAnalysisStarted $n"
  } else {
    fileevent $analysisCoach(pipe$n) readable "::blunder::processAnalysisInput"
    after 1000 "::blunder::checkAnalysisStarted $n"
  }
  
}

# ======================================================================
# ::blunder::closeEngine
#   Close an engine.
# ======================================================================
proc ::blunder::closeEngine {n} {
  global windowsOS analysisCoach
  
  # Check the pipe is not already closed:
  if {$analysisCoach(pipe$n) == ""} {
    set ::coachWin$n 0
    return
  }
  
  # Send interrupt signal if the engine wants it:
  if {(!$windowsOS)  &&  $analysisCoach(send_sigint$n)} {
    catch {exec -- kill -s INT [pid $analysisCoach(pipe$n)]}
  }
  
  # Some engines in analyze mode may not react as expected to "quit"
  # so ensure the engine exits analyze mode first:
  sendToEngine $n "exit"
  sendToEngine $n "quit"
  flush $analysisCoach(pipe$n)
  
  # Uncomment the following line to turn on blocking mode before
  # closing the engine (but probably not a good idea!)
  #   fconfigure $analysisCoach(pipe$n) -blocking 1
  
  # Close the engine, ignoring any errors since nothing can really
  # be done about them anyway -- maybe should alert the user with
  # a message box?
  catch {close $analysisCoach(pipe$n)}
  
  if {$analysisCoach(log$n) != ""} {
    catch {close $analysisCoach(log$n)}
    set analysisCoach(log$n) ""
  }
  set analysisCoach(pipe$n) ""
  set ::analysisWin$n 0
}
# ======================================================================
# sendToEngine:
#   Send a command to a running analysis engine.
# ======================================================================
proc ::blunder::sendToEngine {n text} {
  logEngine $n "Scid  : $text"
  catch {puts $::analysisCoach(pipe$n) $text}
}

# ======================================================================
# checkAnalysisStarted
#   Called a short time after an analysis engine was started
#   to send it commands if Scid has not seen any output from
#   it yet.
# ======================================================================
proc ::blunder::checkAnalysisStarted {n} {
  global analysisCoach
  if {$analysisCoach(seen$n)} { return }
  
  # Some Winboard engines do not issue any output when
  # they start up, so the fileevent above is never triggered.
  # Most, but not all, of these engines will respond in some
  # way once they have received input of some type.  This
  # proc will issue the same initialization commands as
  # those in processAnalysisInput below, but without the need
  # for a triggering fileevent to occur.
  
  ::blunder::logEngineNote $n {Quiet engine (still no output); sending it initial commands.}
  set analysisCoach(seen$n) 1
  ::blunder::sendToEngine $n "xboard"
  ::blunder::sendToEngine $n "protover 2"
  ::blunder::sendToEngine $n "post"
  ::blunder::sendToEngine $n "ponder off"
  
  # Prevent some engines from making an immediate "book"
  # reply move as black when position is sent later:
  ::blunder::sendToEngine $n "force"
}

# ======================================================================
#
# processInput from the engine blundering (Phalanx)
#
# ======================================================================
proc ::blunder::processInput {{n 1}} {
  global analysisCoach analysis
  
  # Get one line from the engine:
  set line [gets $analysisCoach(pipe$n)]
  ::blunder::logEngine $n "Engine: $line"
  
  # Check that the engine did not terminate unexpectedly:
  if {[eof $analysisCoach(pipe$n)]} {
    fileevent $analysisCoach(pipe$n) readable {}
    catch {close $analysisCoach(pipe$n)}
    set analysisCoach(pipe$n) ""
    ::blunder::logEngineNote $n {Engine terminated without warning.}
    catch {destroy .analysisWin$n}
    tk_messageBox -type ok -icon info -parent . -title "Scid" \
        -message "The analysis engine $n terminated without warning; it probably crashed or had an internal error."
  }
  
  if {! $analysisCoach(seen$n)} {
    # First line of output from the program, so send initial commands:
    ::blunder::logEngineNote $n {First line from engine seen; sending it initial commands now.}
    set analysisCoach(seen$n) 1
    ::blunder::sendToEngine $n "xboard"
    ::blunder::sendToEngine $n "post"
  }
  
  ::blunder::makePhalanxMove $line
  
}

# ======================================================================
# processAnalysisInput
#   Called from a fileevent whenever there is a line of input
#   from an analysis engine waiting to be processed.
# ======================================================================
proc ::blunder::processAnalysisInput { } {
  global analysisCoach analysis lscore
  set n 2
  # Get one line from the engine:
  set line [gets $analysisCoach(pipe$n)]
  ::blunder::logEngine $n "Engine: $line"
  
  # Check that the engine did not terminate unexpectedly:
  if {[eof $analysisCoach(pipe$n)]} {
    fileevent $analysisCoach(pipe$n) readable {}
    catch {close $analysisCoach(pipe$n)}
    set analysisCoach(pipe$n) ""
    ::blunder::logEngineNote $n {Engine terminated without warning.}
    catch {destroy .analysisWin$n}
    tk_messageBox -type ok -icon info -parent . -title "Scid" \
        -message "The analysis engine $n terminated without warning; it probably crashed or had an internal error."
  }
  
  if {! $analysisCoach(seen$n)} {
    # First line of output from the program, so send initial commands:
    ::blunder::logEngineNote $n {First line from engine seen; sending it initial commands now.}
    set analysisCoach(seen$n) 1
    ::blunder::sendToEngine $n "xboard"
    ::blunder::sendToEngine $n "protover 2"
    ::blunder::sendToEngine $n "ponder off"
    ::blunder::sendToEngine $n "post"
  }
  
  # Check for "feature" commands so we can determine if the engine
  # has the setboard and analyze commands:
  #
  if {! [string compare [string range $line 0 6] "feature"]} {
    if {[string match "*analyze=1" $line]} { set analysisCoach(has_analyze$n) 1 }
    if {[string match "*setboard=1" $line]} { set analysisCoach(has_setboard$n) 1 }
    if {[string match "*usermove=1" $line]} { set analysisCoach(wants_usermove$n) 1 }
    if {[string match "*sigint=1" $line]} { set analysisCoach(send_sigint$n) 1 }
    
    return
  }
  
  # Check for a line starting with "Crafty", so Scid can work well
  # with older Crafty versions that do not recognize "protover":
  #
  if {! [string compare [string range $line 0 5] "Crafty"]} {
    ::blunder::logEngineNote $n {Seen "Crafty"; assuming analyze and setboard commands.}
    set major 0
    if {[scan $line "Crafty v%d.%d" major minor] == 2  &&  $major >= 18} {
      ::blunder::logEngineNote $n {Crafty version is >= 18.0; assuming scores are from White perspective.}
      set analysisCoach(invertScore$n) 0
    }
    # Turn off crafty logging, to reduce number of junk files:
    ::blunder::sendToEngine $n "log off"
    # Set a fairly low noise value so Crafty is responsive to board changes,
    # but not so low that we get lots of short-ply search data:
    ::blunder::sendToEngine $n "noise 1000"
    set analysisCoach(isCrafty$n) 1
    set analysisCoach(has_setboard$n) 1
    set analysisCoach(has_analyze$n) 1
    return
  }
  
  # Scan the line from the engine for the analysis data:
  #
  set res [scan $line "%d%c %d %d %d %\[^\n\]\n" \
      temp_depth dummy temp_score \
      temp_time temp_nodes temp_moves]
  if {$res == 6} {
    set analysisCoach(score$n) $temp_score
    set analysisCoach(moves$n) [::blunder::formatAnalysisMoves $temp_moves]
    # Convert score to pawns from centipawns:
    set analysisCoach(score$n) [expr double($analysisCoach(score$n)) / 100.0]
    set lscore [lreplace $lscore end end $analysisCoach(score$n)]
    ::blunder::updateAnalysisText
    if {! $analysisCoach(seenEval$n)} {
      # This is the first evaluation line seen, so send the current
      # position details to the engine:
      set analysisCoach(seenEval$n) 1
    }
    return
  }
  
}
# ======================================================================
# formatAnalysisMoves:
#   Given the text at the end of a line of analysis data from an engine,
#   this proc tries to strip out some extra stuff engines add to make
#   the text more compatible for adding as a variation.
# ======================================================================
proc ::blunder::formatAnalysisMoves {text} {
  # Yace puts ".", "t", "t-" or "t+" at the start of its moves text,
  # unless directed not to in its .ini file. Get rid of it:
  if {[strIsPrefix ". " $text]} { set text [string range $text 2 end]}
  if {[strIsPrefix "t " $text]} { set text [string range $text 2 end]}
  if {[strIsPrefix "t- " $text]} { set text [string range $text 3 end]}
  if {[strIsPrefix "t+ " $text]} { set text [string range $text 3 end]}
  
  # Trim any initial or final whitespace:
  set text [string trim $text]
  
  # Yace often adds "H" after a move, e.g. "Bc4H". Remove them:
  regsub -all "H " $text " " text
  
  # Crafty adds "<HT>" for a hash table comment. Change it to "{HT}":
  regsub "<HT>" $text "{HT}" text
  
  return $text
}

# ======================================================================
# startAnalyzeMode:
#   Put the engine in analyze mode
# ======================================================================
proc ::blunder::startAnalyze { { n 2 } } {
  global analysisCoach isLimitedAnalysisTime analysisTime
  
  if { !$analysisCoach(has_analyze$n)} {
    error "No analyze engine"
    return
  }
  # Check that the engine has not already had analyze mode started:
  if {$analysisCoach(analyzeMode$n)} {
    ::blunder::sendToEngine $n "exit"
  }
  
  set analysisCoach(analyzeMode$n) 1
  after cancel ::blunder::stopAnalyze
  
  ::blunder::sendToEngine $n "setboard [sc_pos fen]"
  ::blunder::sendToEngine $n "analyze"
  if { $isLimitedAnalysisTime == 1 }  {
    after [expr 1000 * $analysisTime] ::blunder::stopAnalyze
  }
  
}
# ======================================================================
# stopAnalyzeMode:
#   Stop the engine analyze mode
# ======================================================================
proc ::blunder::stopAnalyze { { n 2 } } {
  global analysisCoach isLimitedAnalysisTime analysisTime
  if { !$analysisCoach(has_analyze$n)} {
    error "No analyze engine"
    return
  }
  # Check that the engine has already had analyze mode started:
  if {!$analysisCoach(analyzeMode$n)} { return }
  set analysisCoach(analyzeMode$n) 0
  ::blunder::sendToEngine $n "exit"
}
# ======================================================================
# phalanxGo
#		it is phalanx's turn to play
# ======================================================================
proc ::blunder::phalanxGo {} {
  global analysisCoach
  
  after cancel ::blunder::phalanxGo
  
  # check if Phalanx is already thinking
  if { $analysisCoach(automoveThinking1) == 1 } {
    after 1000 ::blunder::phalanxGo
    return
  }
  
  if { [sc_pos side] != [::blunder::getPhalanxColor] } {
    after 1000 ::blunder::phalanxGo
    return
  }
  
  # original Phalanx does not have 'setboard' : Pascal Georges added it ...
  set analysisCoach(automoveThinking1) 1
  sendToEngine 1 "setboard [sc_pos fen]"
  sendToEngine 1 "go"
  after 1000 ::blunder::phalanxGo
}
# ======================================================================
# makePhalanxMove:
#
# ======================================================================
proc ::blunder::makePhalanxMove { input } {
  global lscore analysisCoach currentPosHash resignCount resignValue
  
  # The input move is of the form "my move is MOVE"
  if {[scan $input "my move is %s" move] != 1} { return 0 }
  
  # Phalanx will move : update the score list to detect any blunder
  if {$analysisCoach(invertScore2)  && (![string compare [sc_pos side] "black"])} {
    lappend lscore [expr {0.0 - $analysisCoach(score2)} ]
  } else  {
    lappend lscore $analysisCoach(score2)
  }
  
  # if the resignValue has been reached more than 3 times in a raw, resign
  if { ( [getPhalanxColor] == "black" && [lindex $lscore end] > $resignValue ) || \
        ( [getPhalanxColor] == "white" && [lindex $lscore end] < [expr 0.0 - $::resignValue] ) } {
    incr resignCount
  } else  {
    set resignCount 0
  }
  
  # check the sequence of moves
  # in case of any event (board setup, move back/forward), reset score list
  if { ![sc_pos isAt start] && ![sc_pos isAt vstart]} {
    sc_move back 1
      if { [sc_pos hash] != $currentPosHash} {
        set lscore {}
        updateAnalysisText
      }
    sc_move forward 1    
  } else  {
    if { [sc_pos hash] != $currentPosHash} {
      set lscore {}
      updateAnalysisText
    }
  }
  
  # play the move
  if {[catch {sc_move addSan $move}]} {
    # No move from Phalanx : remove the score (last element)
    set lscore [lreplace $lscore end end]
    return 0
  }
  
  set analysisCoach(automoveThinking1) 0
  set currentPosHash [sc_pos hash]
  
  # Phalanx has moved : start coach analysis
  if {$analysisCoach(seen2)} {
    ::blunder::startAnalyze
  }
  
  updateBoard -pgn
  
  if { $resignCount > 3 } {
    tk_messageBox -type ok -message "I resign" -parent .board -icon info
    set resignCount 0
  }
  
  return 1
}

# ======================================================================
# updateAnalysisText
#   Update the text in an analysis window.
# ======================================================================
proc ::blunder::updateAnalysisText { } {
  global analysisCoach showblunder blunderWarningLabel \
      showblunder showblundervalue showblunderfound showmovevalue showevaluation lscore threshold \
      lastblundervalue prev_lastblundervalue scoreLabel blunderpending prev_blunderpending \
      blunder::threshold1 blunder::threshold2 blunder::threshold3 blunder::thresholdGameWasDeadAnyway
  
  # There are less than 2 scores in the list
  if {[llength $lscore] < 2} {
    set blunderWarningLabel $::tr(Noinfo)
    set scoreLabel ""
    if {[llength $lscore] == 1 && $showevaluation } {
      set scoreLabel "Score : [lindex $lscore end]"
    }
    return
  }
  
  set sc1 [lindex $lscore end]
  set sc2 [lindex $lscore end-1]
  
  if { $analysisCoach(automoveThinking1) } {
    set blunderWarningLabel $::tr(Noinfo)
  }
  
  # Check if a blunder was made by Phalanx at last move.
  # The check is done during player's turn
  if { $showblunder && [::blunder::getPhalanxColor] != [sc_pos side] } {
    if {[llength $lscore] >=2} {
      if { ($sc1 - $sc2 > $threshold && [::blunder::getPhalanxColor] == "black") || \
            ($sc1 - $sc2 < [expr 0.0 - $threshold] && [::blunder::getPhalanxColor] == "white") } {
        set lastblundervalue [expr $sc1-$sc2]
        # append a ?!, ? or ?? to the move if there is none yet and if the game was not dead yet
        # (that is if the score was -6, if it goes down to -10, this is a normal evolution
        if { [expr abs($sc2)] < $thresholdGameWasDeadAnyway } {
          sc_pos clearNags
          set b [expr abs($lastblundervalue)]
          if { $b >= $threshold1 && $b < $threshold2 } {
            sc_pos addNag "?!"
          } elseif { $b >= $threshold2 && $b < $threshold3 }  {
            sc_pos addNag "?"
          } elseif  { $b >= $threshold3 } {
            sc_pos addNag "??"
          }
        }
        
        .coachWin.finformations.l1 configure -bg LightCoral
        if { $showblundervalue } {
          set tmp $::tr(blunder)
          append tmp [format " %+8.2f" [expr abs($sc1-$sc2)]]
          set blunderWarningLabel $tmp
          set blunderpending 1
        } else {
          set blunderWarningLabel "$::tr(blunder) !"
        }
      } else {
        sc_pos clearNags
        .coachWin.finformations.l1 configure -bg linen
        set blunderWarningLabel $::tr(Noblunder)
        set blunderpending 0
      }
    }
  } else {
    set blunderWarningLabel "---"
  }
  
  if { !$showblunder || $analysisCoach(automoveThinking1) } {
    set blunderWarningLabel "---"
  }
  
  # displays current score sent by the "good" engine (Crafty)
  if { $showevaluation } {
    set scoreLabel "Score : $sc1"
  } else {
    set scoreLabel ""
  }
}

# ======================================================================
# getPhalanxColor
#   Returns "white" or "black" (Phalanx always plays at top)
# ======================================================================
proc ::blunder::getPhalanxColor {} {
  # Phalanx always plays for the upper side
  if { [::board::isFlipped .board] == 0 } {
    return "black"
  } else  {
    return "white"
  }
}

# ======================================================================
# logEngine:
#   Log Scid-Engine communication.
# ======================================================================
proc ::blunder::logEngine {n text} {
  global analysisCoach
  
  # Print the log message to  if applicable:
  if {$::analysisCoach(log_stdout)} {
    puts stdout "$n $text"
  }
  
  if {$::analysisCoach(log$n) != ""} {
    puts $::analysisCoach(log$n) $text
    
    # Close the log file if the limit is reached:
    incr analysisCoach(logCount$n)
    if {$analysisCoach(logCount$n) >= $analysisCoach(logMax)} {
      puts $::analysisCoach(log$n) \
          "NOTE  : Log file size limit reached; closing log file."
      catch {close $analysisCoach(log$n)}
      set analysisCoach(log$n) ""
    }
  }
}

# ======================================================================
# logEngineNote:
#   Add a note to the engine communication log file.
# ======================================================================
proc ::blunder::logEngineNote {n text} {
  logEngine $n "NOTE  : $text"
}

###
### End of file: blunder.tcl
###
